<?php
namespace bhl_bind;
use Exception;

function get_twig(array $inc_path = [])
{
  array_unshift($inc_path, __DIR__ . "/../tpl/");
  $loader = new \Twig\Loader\FilesystemLoader($inc_path);

  $twig = new \Twig\Environment($loader, [
    'debug' => true, 
    'autoescape' => false, 
    'strict_variables' => true]
  );
  $twig->addExtension(new \Twig\Extension\DebugExtension());

  add_twig_support($twig);

  return $twig;
}

function supported_tokens()
{
  return [
    'POD',
    'default',
    'bhl_bind',
    'bhl_no_new',
    'bhl_ecs_component',
    'bhl_ecs_component_ref',
    'bhl_coroutine',
    'bhl_static',
    'bhl_native_class',
    'bhl_native_class_params',
    'bhl_native_prefix',
    'bhl_custom_rw',
    'bhl_get',
    'bhl_set',
    'bhl_ref_arg',
    'bhl_bin_op'
  ];
}

function prepare_meta(\mtgMetaInfo $orig)
{
  //NOTE: making a deep clone
  $meta = unserialize(serialize($orig));

  $arr_proxies = array();
  foreach($meta->getUnits() as $u)
  {
    if($u->object instanceof \mtgMetaStruct && !$u->object->hasToken('bhl_ecs_component'))
    {
      $need_to_replace = false;
      $fields = $u->object->getFields(); 
      //NOTE: let's replace arrays with array proxies
      foreach($fields as $name => $fld)
      {
        if($fld->getType() instanceof \mtgArrType)
        {
          $proxy_name = "List_{$fld->getType()->getValue()->getName()}";
          if(!isset($arr_proxies[$proxy_name]))
          {
            $proxy = new \mtgMetaStruct($proxy_name);
            $proxy->setToken('bhl_native_arr_proxy', $fld->getType()->getValue());
            $meta->addUnit(new \mtgMetaInfoUnit('', $proxy));
            $arr_proxies[$proxy_name] = $proxy;
          }
          else
            $proxy = $arr_proxies[$proxy_name];

          $new_fld = new \mtgMetaField($name, new \mtgTypeRef($proxy)); 
          //forcing no setter
          $new_fld->setToken('bhl_set', 0);
          $fields[$name] = $new_fld;

          $need_to_replace = true;
        }
      }

      if($need_to_replace)
      {
        foreach($u->object->getFields() as $fld)
          $u->object->delField($fld);
        $u->object->setFields($fields);
      }
    }
  }

  return $meta;
}

function add_twig_support(\Twig\Environment $twig)
{
  $twig->addTest(new \Twig\TwigTest('instanceof', 
    function($obj, $class) 
    { 
      return (new \ReflectionClass($class))->isInstance($obj);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('ns_last', 
    function($str) 
    { 
      return ns_last($str);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('ns_prefix', 
    function($str) 
    { 
      return ns_prefix($str);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('bhl_type', 
    function($type) 
    { 
      return bhl_type($type);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('bhl_type_ref', 
    function($type) 
    { 
      if(is_string($type))
        return "types.T(\"$type\")";

      return bhl_type_ref($type);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('native_type', 
    function($type) 
    { 
      return native_type($type);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('norm_name', 
    function($name) 
    { 
      return norm_name($name);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('rtrim_comma', 
    function($str) 
    { 
      return rtrim(rtrim($str), ',');
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('replace_in', 
    function($txt, array $args = []) 
    { 
      return str_replace($args[1], $txt, $args[0]);
    },
    ['is_variadic' => true]
  ));
  $twig->addFunction(new \Twig\TwigFunction('Error', 
    function($e) 
    {
      throw new Exception($e);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('Warn', 
    function($e) 
    {
      Warn($e);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('has_token', 
    function($o, $token) 
    { 
      return $o->hasToken($token);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('token', 
    function($o, $name) 
    {
      return $o->getToken($name);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('token_or', 
    function($o, $name, $v) 
    {
      if($o->hasToken($name))
        return $o->getToken($name);
      else
        return $v;
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('count_required_args', 
    function(\mtgMetaFunc $fn) 
    {
      $req_args_num = 0;
      foreach($fn->getArgs() as $arg)
      {
        if($arg->hasToken('default'))
          break;
        ++$req_args_num;
      }
      return $req_args_num;
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('count_default_args', 
    function(\mtgMetaFunc $fn) 
    {
      $def_args_num = 0;
      foreach($fn->getArgs() as $arg)
      {
        if($arg->hasToken('default'))
          ++$def_args_num;
        else if($def_args_num > 0)
          throw new Exception("Field '{$name}' in {$fn->getName()} is expected to be tagged with @default");
      }
      return $def_args_num;
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('types_array', 
    function(\mtgType $type) 
    {
      $arr = array();
      if($type instanceof \mtgMultiType)
        $arr = $type->getValues();
      else
        $arr[] = $type;
      return $arr;
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('slice_units', 
  function(array $units, $max) 
  { 
    $pages = paginate(sizeof($units), $max);
   
    $sliced = array();

    $units_keys = array_keys($units);

    foreach($pages as $idx => $page)
    {
      $slice = array();
      for($i = $page[0];$i<$page[1];++$i)
      {
        $slice[] = $units[$units_keys[$i]];
      }
      $sliced[$idx] = $slice;  
    }
    return $sliced;
  }
  ));
}

function Warn($msg)
{
  //NOTE: we can't just echo the value because it will
  //      be rendered into resulting code
  register_shutdown_function(function() use($msg) {
    echo "BHL BIND WARN: $msg\n";
  });
}

function paginate($total, $step)
{
  //pages are returned as an array where each element is in interval [N, Y)
  $pages = array();

  $steps = (int)($total/$step);
  $rest = $total % $step;

  for($i=1;$i<=$steps;++$i)
    $pages[] = array(($i-1)*$step, $i*$step);

  if($rest != 0)
    $pages[]  = array(($i-1)*$step, ($i-1)*$step + $rest);

  return $pages;
}

function ns_last($name)
{
  $items = explode('.', $name);
  return end($items);
}

function ns_prefix($name)
{
  $items = explode('.', $name);
  if(sizeof($items) == 1)
    return '';

  $prefix = '';
  for($i=0;$i<sizeof($items)-1;++$i)
    $prefix .= 'Nest("'.$items[$i].'").';

  return $prefix;
}

function norm_name($name)
{
  return str_replace('.', '__', $name);
}

function bhl_type_ref(\mtgType $type = null)
{
  if($type === null)
    return 'types.T("void")';

  if($type instanceof \mtgBuiltinType)
  {
    if($type->isFloat() || $type->isDouble())
      return 'types.T("float")';
    else if($type->isInt() || $type->isUint())
      return 'types.T("int")';
  }
  else if($type instanceof \mtgMetaFunc)
  {
    $is_coro = $type->hasToken("bhl_coroutine");
    $str = '';
    $str .= "types.TFunc(".($is_coro?"true":"false").", ";
    $str .= ($type->getReturnType() ? bhl_type_ref($type->getReturnType()) : bhl_type_ref(null)) . ",";
    foreach($type->getArgs() as $arg)
      $str .= bhl_type_ref($arg->getType()) . ",";
    $str = rtrim($str, ',');
    $str .= ")";

    return $str;
  }
  else if($type instanceof \mtgMultiType)
  {
    $str = 'types.T(';
    foreach($type->getValues() as $sub_type)
      $str .= bhl_type_ref($sub_type) . ",";
    $str = rtrim($str, ",");
    $str .= ")";

    return $str;
  }
  else if($type instanceof \mtgArrType)
  {
    return 'types.TArr(' . bhl_type_ref($type->getValue()) . ')';
  }
  
  //fallback
  return 'types.T("' . $type. '")';
}

function bhl_type(\mtgType $type = null)
{
  if($type === null)
    return "void";

  if($type instanceof \mtgBuiltinType)
  {
    if($type->isFloat() || $type->isDouble())
      return "float";
    else if($type->isInt() || $type->isUint())
      return "int";
  }
  else if($type instanceof \mtgMetaFunc)
  {
    $ret_type = $type->getReturnType() ? bhl_gen_type($type->getReturnType()) : 'void';

    $str = "";
    $str .= "func $ret_type(";
    foreach($type->getArgs() as $arg)
      $str .= bhl_type($arg->getType()) . ",";
    $str = rtrim($str, ',');
    $str .= ")";
    return $str;
  }
  else if($type instanceof \mtgMultiType)
  {
    $str = '';
    foreach($type->getValues() as $sub_type)
      $str .= bhl_type($sub_type) . ",";
    return rtrim($str, ",");
  }
  else if($type instanceof \mtgArrType)
    return "[]" . bhl_type($type->getValue());

  //fallback
  return ''.$type;
}

function native_type(\mtgType $type)
{
  //any special case
  if($type == "any")
    return "bhl.Val";

  if($type instanceof \mtgArrType)
  {
    if($type->getValue()->getName() == "any")
      return "bhl.ValList";
    else
      return "HList<".native_type($type->getValue()).">"; 
  }

  //func special case
  if($type instanceof \mtgMetaFunc)
    return "VM.FuncPtr";

  $alias = null;
  if($type instanceof \mtgUserType)
    $alias = $type->getToken('bhl_native_class');

  if($type instanceof \mtgArrType)
    $orig_name = $type->getValue()->getName();
  else
    $orig_name = $type->getName();
  if($alias)
    $type->setName($alias);

  $ntype = cs_type($type);

  //TODO: looks suspicious
  if($alias)
    $type->setName($orig_name);

  return $ntype;
}

function cs_type(\mtgType $type)
{
  if($type instanceof \mtgBuiltinType || $type instanceof \mtgUserType)
    return cs_simple_type($type);
  else if($type instanceof \mtgArrType)
    return "List<" . cs_simple_type($type->getValue()) . ">";
  else
    throw new Exception("Unknown type '{$type}'");
}

function cs_simple_type(\mtgType $type)
{
  if($type instanceof \mtgBuiltinType)
  {
    switch($type->getName())
    {
      case "int8":
        return "sbyte";
      case "uint8":
        return "byte";
      case "int16":
        return "short";
      case "uint16":
        return "ushort";
      case "int32":
      case "int":
        return "int";
      case "uint32":
      case "uint":
        return "uint";
      case "float":
        return "float";
      case "double":
        return "double";
      case "uint64":
        return "ulong";
      case "int64":
        return "long";
      case "string":
        return "string";
      case "bool":
        return "bool";
      case "blob":
        return "byte[]";
    }
    throw new Exception("Unknown type '{$type}'");
  }
  return $type->getName();
}

